import re


class TBLine:
    def __init__(self, tb_line: list):
        self.balance_type = tb_line[0]
        self.entity_code = tb_line[1]
        self.entity_name = tb_line[2]
        self.month = tb_line[3]
        self.year = tb_line[4]
        self.account_code = tb_line[5]
        self.account_name = tb_line[6]
        self.currency_code = tb_line[7]
        self.currency_name = tb_line[8]
        self.begin_balance_ori_currency = tb_line[9]
        self.year_debit_ori_currency = tb_line[10]
        self.year_credit_ori_currency = tb_line[11]
        self.period_debit_ori_currency = tb_line[12]
        self.period_credit_ori_currency = tb_line[13]
        self.end_balance_ori_currency = tb_line[14]
        self.begin_balance_local_currency = tb_line[15]
        self.year_debit_local_currency = tb_line[16]
        self.year_credit_local_currency = tb_line[17]
        self.period_debit_local_currency = tb_line[18]
        self.period_credit_local_currency = tb_line[19]
        self.end_balance_local_currency = tb_line[20]
        self.begin_balance_report_currency = tb_line[21]
        self.year_debit_report_currency = tb_line[22]
        self.year_credit_report_currency = tb_line[23]
        self.period_debit_report_currency = tb_line[24]
        self.period_credit_report_currency = tb_line[25]
        self.end_balance_report_currency = tb_line[26]
        self.year_pl_ori_currency = tb_line[27]
        self.period_pl_ori_currency = tb_line[28]
        self.year_pl_local_currency = tb_line[29]
        self.period_pl_local_currency = tb_line[30]
        self.year_pl_report_currency = tb_line[31]
        self.period_pl_report_currency = tb_line[32]
        self.begin_qty = tb_line[33]
        self.year_debit_qty = tb_line[34]
        self.year_credit_qty = tb_line[35]
        self.period_debit_qty = tb_line[36]
        self.period_credit_qty = tb_line[37]
        self.end_qty = tb_line[38]
        self.asst_code = tb_line[39]
        self.asst_name = tb_line[40]
        self.has_asst = tb_line[41]
        self.has_asst_bool = True if tb_line[41] == '1' else False


class AsstTBLine:
    def __init__(self, tb_line: list):
        self.balance_type = tb_line[0]
        self.entity_code = tb_line[1]
        self.entity_name = tb_line[2]
        self.month = tb_line[3]
        self.year = tb_line[4]
        self.account_code = tb_line[5]
        self.account_name = tb_line[6]
        self.currency_code = tb_line[7]
        self.currency_name = tb_line[8]
        self.begin_balance_ori_currency = tb_line[9]
        self.year_debit_ori_currency = tb_line[10]
        self.year_credit_ori_currency = tb_line[11]
        self.period_debit_ori_currency = tb_line[12]
        self.period_credit_ori_currency = tb_line[13]
        self.end_balance_ori_currency = tb_line[14]
        self.begin_balance_local_currency = tb_line[15]
        self.year_debit_local_currency = tb_line[16]
        self.year_credit_local_currency = tb_line[17]
        self.period_debit_local_currency = tb_line[18]
        self.period_credit_local_currency = tb_line[19]
        self.end_balance_local_currency = tb_line[20]
        self.begin_balance_report_currency = tb_line[21]
        self.year_debit_report_currency = tb_line[22]
        self.year_credit_report_currency = tb_line[23]
        self.period_debit_report_currency = tb_line[24]
        self.period_credit_report_currency = tb_line[25]
        self.end_balance_report_currency = tb_line[26]
        self.year_pl_ori_currency = tb_line[27]
        self.period_pl_ori_currency = tb_line[28]
        self.year_pl_local_currency = tb_line[29]
        self.period_pl_local_currency = tb_line[30]
        self.year_pl_report_currency = tb_line[31]
        self.period_pl_report_currency = tb_line[32]
        self.begin_qty = tb_line[33]
        self.year_debit_qty = tb_line[34]
        self.year_credit_qty = tb_line[35]
        self.period_debit_qty = tb_line[36]
        self.period_credit_qty = tb_line[37]
        self.end_qty = tb_line[38]
        self.asst_code = tb_line[39]
        self.asst_name = tb_line[40]
        self.asst_full_desc = tb_line[41]
        self.asst_full_name = tb_line[42]
        self.asst_full_code = tb_line[43]
        # parse asst information
        asst_type_list = self.asst_full_code.split(';')
        self.asst_type1 = asst_type_list[0].split('_!')[1].split(':')[0]
        self.asst_code1 = asst_type_list[0].split('_!')[1].split(':')[1]
        try:
            self.asst_type2 = asst_type_list[1].split('_!')[1].split(':')[0]
            self.asst_code2 = asst_type_list[1].split('_!')[1].split(':')[1]
        except IndexError:
            self.asst_type2 = None
            self.asst_code2 = None

    def get_main_info_str(self):
        info = ' '.join(['balance_type', self.balance_type,
                         'currency_code', self.currency_code,
                         'account_code', self.account_code,
                         'asst_type', self.asst_type1,
                         'asst_code', self.asst_code1,
                         'year_end_balance', self.end_balance_local_currency])
        return info


class AsstTB:
    def __init__(self, tb: list):
        self.asst_list = []
        if tb:
            for each in tb:
                if not re.match('^\\d', str(each[0])):
                    continue
                self.asst_list.append(AsstTBLine(each))


class TB:
    def __init__(self, tb: list):
        """
        tested the size of one yearly tb, with all asst accounts info, the size was approximately 24.4M
        which should be considered acceptable
        :param tb: the original result from kd webservice
        """
        self.record_list = []
        self.asst_tb_dict = {}
        if tb:
            for line in tb:
                if not re.match('^\\d', str(line[0])):
                    continue
                self.record_list.append(TBLine(line))

    def is_leaf_account(self, account_code) -> bool:
        for each in self.record_list:
            if re.match(account_code + '.+', each.account_code):
                return False
        return True


def seperate_voucher_list(voucher_list: list, renumber=False) -> dict:
    """
    clean the voucher list
    :param voucher_list: use voucher number and company number as unique key to identify voucher,
     the key to result dictionary is 'voucherNumber_companyNumber'
    :param renumber: WARNING! if this args is True,
     the function will comb the list and renumber the sequence of entry, which may lead unknown error.
    :return: the dict of voucher list, the key should be the voucher number
    """
    seperated_voucher_dict = {}
    for voucher in voucher_list:
        voucher_key = '_'.join([str(voucher['voucherNumber']), str(voucher['companyNumber'])])
        if voucher_key not in seperated_voucher_dict:
            seperated_voucher_dict.setdefault(voucher_key, [])
        seperated_voucher_dict[voucher_key].append(voucher)
    if renumber:
        for voucher_key in seperated_voucher_dict:
            for new_entry_int, voucher in enumerate(seperated_voucher_dict[voucher_key]):
                voucher['entrySeq'] = new_entry_int + 1
    return seperated_voucher_dict


def _get_standard_voucher(assit=0):
    standard_voucher = {
        'companyNumber': '',
        'bookedDate': '',
        'bizDate': '',
        'periodYear': 0,
        'periodNumber': 0,
        'voucherType': '',
        'voucherNumber': '',
        'entrySeq': 0,
        'voucherAbstract': '',
        'accountNumber': '',
        'currencyNumber': '',
        'localRate': 1,
        'entryDC': 0,
        'originalAmount': 0.0,
        'debitAmount': 0.0,
        'creditAmount': 0.0,
        'creator': '',
    }
    if assit == 1:
        standard_voucher['asstActType1'] = ''
        standard_voucher['asstActNumber1'] = ''
    elif assit == 2:
        standard_voucher['asstActType1'] = ''
        standard_voucher['asstActNumber1'] = ''
        standard_voucher['asstActType2'] = ''
        standard_voucher['asstActNumber2'] = ''
    return standard_voucher



import codecs
import os
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()


VERSION = '0.1.0'
DESCRIPTION = '金蝶云平台凭证批量处理封装工具。因为依赖包古老，因此需要注意降级pip和setuptools版本，建议降级到pip-20.1.1, setuptools-57.5.0'

setup(
    name="KDVoucherUtil",
    version=VERSION,
    author="Arthur Yu",
    author_email="arthurzero@homtail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    license='MIT',
    install_requires=[
        "suds-jurko>=0.6",
    ],
    # keywords=['python', 'menu', 'dumb_menu','windows','mac','linux'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



import pickle
import os
import re
import copy

from suds.client import Client
from suds.sax.element import Element
from . import *


class VoucherWorker:
    def __init__(self, domain, sln_name, dc_name, language, db_type, auth_pattern):
        self.domain = domain
        self.sln_name = sln_name
        self.dc_name = dc_name
        self.language = language
        self.db_type = db_type
        self.auth_pattern = auth_pattern
        self.ssid = None

    def login(self, user_name, user_pwd) -> str:
        try:
            userName = user_name
            password = user_pwd
            slnName = self.sln_name
            dcName = self.dc_name
            language = self.language
            dbType = self.db_type
            authPattern = self.auth_pattern
            client = Client(self.domain + '/ormrpc/services/EASLogin?wsdl')
            result = client.service.login(userName, password, slnName, dcName, language, dbType, authPattern)
            ssid = re.search('sessionId = "' + '(.*?)' + '"', str(result)).group(1)
            print('webservice login successfully， login as ' + userName)
            self.ssid = ssid
            return ssid
        except Exception as e:
            print(e)
            return 'Login Error - original error message: \n' + str(e)

    def import_voucher(self, voucher_list: list, ssid=None) -> dict:
        """
        if an ssid is not given, the worker will use the previous ssid to try to import voucher
        WARNING: the rollback function of kingdae service seems to be unreliable, STRONGLY RECOMMEND to import one voucher at a time!!!
        :param voucher_list: standard voucher please refer to VoucherUtil - get_standard_voucher
        :param ssid:
        :return: kingdae returns a a list, which is in form of ["error code||voucher type||year||month||chs result||voucher number||result code"]
                the function result should be in form of :
                {
                    'result': str,
                    'type': str,
                    'year': str,
                    'month: str,
                    'error_msg': str
                    'voucher_number': str,
                    'origin_result_str': str,
                    'desc': str,
                }
        """
        if not ssid:
            ssid = self.ssid
        if not ssid:
            raise NotLoginException
        try:
            client = Client(self.domain + '/ormrpc/services/WSGLWebServiceFacade?wsdl')
            client.set_options(timeout=300)
            ssn = Element('SessionId').setText(ssid).setPrefix(p='ns1', u='http://login.webservice.bos.kingdee.com')
            client.set_options(soapheaders=ssn)
            result = client.service.importVoucher(voucher_list, 1, 0, 0)
            parse_result = result[0].split('||')
            return {
                'result': parse_result[0],
                'type': parse_result[1],
                'year': parse_result[2],
                'month': parse_result[3],
                'invoke_msg': parse_result[4],
                'voucher_number': parse_result[5],
                'origin_result_str': result[0],
            }
        except Exception as e:
            return {
                'result': '-2',
                'type': None,
                'year': None,
                'month': None,
                'invoke_msg': None,
                'voucher_number': None,
                'origin_result_str': str(e),
            }

    def delete_voucher(self, entity_code: str, period: str, voucher_number: str, desc: str, ssid=None) -> int:
        if not ssid:
            ssid = self.ssid
        if not ssid:
            raise NotLoginException
        try:
            client = Client(self.domain + '/ormrpc/services/WSGLWebServiceFacade?wsdl')
            client.set_options(timeout=300)
            ssn = Element('SessionId').setText(ssid).setPrefix(p='ns1', u='http://login.webservice.bos.kingdee.com')
            client.set_options(soapheaders=ssn)
            result = client.service.deleteVoucher(entity_code, period, voucher_number, desc)
        except Exception as e:
            result = -1
            print(e)
        return int(result)

    def get_account_balance(self, entity_code: str, str_year: str, str_month: str, ssid=None) -> list:
        """
        the kingdea server will only return up to 1000 records once at a time,
        this function will automatically loop and get every balance of certain period
        :param entity_code:
        :param str_year:
        :param str_month:
        :param ssid:
        :return:
        """
        if not ssid:
            ssid = self.ssid
        if not ssid:
            raise NotLoginException
        result = []
        try:
            client = Client(self.domain + '/ormrpc/services/WSGLWebServiceFacade?wsdl')
            client.set_options(timeout=300)
            ssn = Element('SessionId').setText(ssid).setPrefix(p='ns1', u='http://login.webservice.bos.kingdee.com')
            client.set_options(soapheaders=ssn)
            len_of_tb = 1001
            loop = 0
            result = []
            while len_of_tb == 1001:
                curr_result = client.service.getAccountBalance(entity_code, str(str_year), str(str_month),
                                                               loop * len_of_tb, loop * len_of_tb + len_of_tb)
                result += curr_result if curr_result is not None else []
                len_of_tb = len(curr_result) if curr_result is not None else 0
                loop += 1
            print('Invoke getAccountBalance interface successfully, param: entity_code' + entity_code + '; period: '
                  + str(str_year) + '-' + str(str_month) + '; data length: ' + str(len(result) if result else 0))
        except ConnectionResetError:
            self.get_account_balance(entity_code, str_year, str_month, ssid)
        except Exception as e:
            result = [e]
            print('Unexpected error occurred in invoking getAccountBalance interface, param: ' + entity_code + '; period: '
                  + str(str_year) + '-' + str(str_month))
            print(e)
        finally:
            return result

    def get_assit_balance(self, entity_code: str, account_number: str, str_year: str, str_month: str, ssid: str) -> list:
        """
        the kingdea server will only return up to 1000 records once at a time,
        this function did not consider the case which asst balance longer than 1k
        todo: update the loop function
        :param entity_code:
        :param account_number:
        :param str_year:
        :param str_month:
        :param ssid:
        :return:
        """
        if not ssid:
            ssid = self.ssid
        if not ssid:
            raise NotLoginException
        result = []
        try:
            client = Client(self.domain + '/ormrpc/services/WSGLWebServiceFacade?wsdl')
            client.set_options(timeout=300)
            ssn = Element('SessionId').setText(ssid).setPrefix(p='ns1', u='http://login.webservice.bos.kingdee.com')
            client.set_options(soapheaders=ssn)
            # no loop designed, assit ledger is not supposed to be longer than 1k
            result = client.service.getAssitBalance(entity_code, account_number, str(str_year), str(str_month), 0, 0)
            print('Invoke getAsstBalance interface successfully, param: entity_code:' + entity_code +
                  ';account_code:' + str(account_number) + '; period: '
                  + str(str_year) + '-' + str(str_month) + '; data length: ' + str(len(result) if result else 0))
        except ConnectionResetError:
            self.get_assit_balance(entity_code, account_number, str_year, str_month, ssid)
        except Exception as e:
            result = [e]
            print('Unexpected error occurred in invoking getAsstBalance interface, param: ' + entity_code +
                  ';account_code: ' + str(account_number) + '; period: '
                  + str(str_year) + '-' + str(str_month))
            print(e)
        finally:
            return result

